#include "BasisUtils.h"
#include "Solve.h"
#include "math.h"

inline int powi(int a, int b)
// cheap & nasty a^b calculator.
{
	int result = a;
	
	while (b-- > 1)
		result *= a;
	return(result);
}

/* 
	Calculates the integration matrix for a particular function basis and
	a particular (2D) quadrature rule. 
*/

void MakeBasisMatrix(FunctionBasis &b, const QuadRule &rule, Matd &B)
{
	Int quadPoints = rule.Rows();
	Int i, j;

	B.SetSize(quadPoints, b.Dim());
	
	for (i = 0; i < quadPoints; i++)	// Form B by sampling ...
		for (j = 0; j < b.Dim(); j++)
			B[i][j] = b.Eval(j, sub(rule[i], 1, b.FuncDim()));
}

//	Calculates weight matrix for a rule.

void MakeWeightMatrix(const QuadRule &rule, Matd &W)
{
	Int quadPoints = rule.Rows();
	Int i;
	
	W.SetSize(quadPoints, quadPoints);
	W = vl_zero;

	diag(W, 0) = col(rule, 0);
}

void FindConversionMatrix(FunctionBasis &bFrom, FunctionBasis &bTo, const QuadRule &rule,
	Matd &C)
{
	Matd W, B1, B2, D;
	
	MakeBasisMatrix(bFrom, rule, B1);	
	MakeWeightMatrix(rule, W);	
	MakeBasisMatrix(bTo, rule, B2);

	D = trans(B2) * W * B2;
	C = clamped(inv(D) * trans(B2) * W * B1);
}


// --- Wavelet routines ---------------------------------

void MakeSubBasisMatrix(FunctionBasis &b, const QuadRule &rule, Int child, Int order, Matd &B)
// Takes a function basis, and returns the matrix basis covering a sub-region of it.
{
	Int		quadPoints = rule.Rows();
	Int		i, j;
	Vecd	sx;
	
	B.SetSize(quadPoints, b.Dim());
	
	for (i = 0; i < quadPoints; i++)	// Form B by sampling ...
	{
		sx = b.ChildMap(sub(rule[i], 1, b.FuncDim()), child, order);
		for (j = 0; j < b.Dim(); j++)
			B[i][j] = b.Eval(j, sx);
	}
}

// --- Flatlets -----------------------

Matd my_inv(const Matd &m)
{
        Assert(m.IsSquare(), "(inv) Matrix not square");

    Int                         i, j, k;
    Int                         n = m.Rows();
    Real                      t, pivot, det;
    Real                        max;
    Matd                        A(m);
    Matd                        B(n, n, vl_I);          

    // ---------- Forward elimination ---------- 
        
    det = vl_1;
    
    for (i = 0; i < n; i++)                             // Eliminate in column i, below diag
    {           
                max = -1.0;
                
                for (k = i; k < n; k++)                 // Find a pivot for column i 
                    if (len(A[k][i]) > max)
                    {
                                max = len(A[k][i]);
                                j = k;
                    }
                    
                Assert(max > 0.0, "(inv) Matrix not invertible");
                                        
                if (j != i)                                             // Swap rows i and j
                {                       
                    for (k = i; k < n; k++)
                                Swap(A.Elt(i, k), A.Elt(j, k));
                    for (k = 0; k < n; k++)
                                Swap(B.Elt(i, k), B.Elt(j, k));
                                
                    det = -det;
                }
                
                pivot = A.Elt(i, i);
                Assert(abs(pivot) > 1e-30, "(my_inv) Matrix not invertible");
                det *= pivot;
                
                for (k = i + 1; k < n; k++)             // Only do elements to the right of the pivot 
                    A.Elt(i, k) /= pivot;
                    
                for (k = 0; k < n; k++)
                    B.Elt(i, k) /= pivot;
                    
                // We know that A(i, i) will be set to 1, so don't bother to do it
        
                for (j = i + 1; j < n; j++)
                {                                                               // Eliminate in rows below i 
                    t = A.Elt(j, i);                    // We're gonna zero this guy 
                    for (k = i + 1; k < n; k++) // Subtract scaled row i from row j 
                                A.Elt(j, k) -= A.Elt(i, k) * t; // (Ignore k <= i, we know they're 0) 
                    for (k = 0; k < n; k++)
                                B.Elt(j, k) -= B.Elt(i, k) * t;
                }
    }

    // ---------- Backward elimination ---------- 

    for (i = n - 1; i > 0; i--)                 // Eliminate in column i, above diag 
    {           
                for (j = 0; j < i; j++)                 // Eliminate in rows above i 
                {               
                t = A.Elt(j, i);                        // We're gonna zero this guy 
                for (k = 0; k < n; k++)         // Subtract scaled row i from row j 
                                B.Elt(j, k) -= B.Elt(i, k) * t;
                }
    }
    cout << "det= " << det << endl;
    return(B);
}

void MakeFlatletMatrices(
		FunctionBasis	&basis,
		const QuadRule	&rule,
		FunctionBasis	&flatBasis,
		Int				order,		// How many panes in a dimension. (2 for F2, 3 for F3 etc.)
		Matd			&F,			// basis to flatlet
		Matd			&G			// flatlet to basis
	)
// Returns matrices to convert from a source basis 'basis' to a destination
// flatlet-type basis, made up of (independent) panes of 'flatBasis'.
// F converts from the source to the flatlet basis, and G converts back again.

// Most commonly used to convert from Multiwavelets to Flatlets and back.
// (E.g., An F2 pull can be accomplished by transforming to M2, doing an M2 pull,
// and transforming back to F2.)
// flatBasis is usually a Haar type basis, but doesn't have to be.
{
	Matd	B1, B2, D1, D2, W;
	Int		i, n = powi(order, basis.FuncDim()); 	// n ^ basis.FuncDim().

	F.SetSize(n, basis.Dim());
	G.SetSize(basis.Dim(), n);

	MakeBasisMatrix(flatBasis, rule, B2);
	MakeWeightMatrix(rule, W);
	D2 = clamped(inv(trans(B2) * W * B2));
	MakeBasisMatrix(basis, rule, B1);
	D1 = clamped(inv(trans(B1) * W * B1));

	// each flatlet pane corresponds to a row of F
	// each flatlet pane corresponds to a column of G
	
	for (i = 0; i < n; i++)
	{
		// Make basis matrix for the sub part of the source basis
		MakeSubBasisMatrix(basis, rule, i, order, B1);

		sub(F, i * 1, 0, 1, F.Cols()) = clamped(D2 * trans(B2) * W * B1);
//		sub(G, 0, i * 1, G.Rows(), 1) = clamped(D1 * trans(B1) * W * B2) / Real(n);
	}

	G = inv(trans(F) * F) * trans(F);	// left pseudo-inverse
//	G = trans(F) * inv(F * trans(F));	// right pseudo-inverse
}

