#include "Tensor.h"
#include "Solve.h"

// --- Tensor matrix routines ---------------------------

Bool MatchMatrix(SubMatd m1, SubMatd m2, Real &scale);

Bool MatchMatrix(SubMatd m1, SubMatd m2, Real &scale)
// returns true if m2 is a scalar multiple of m1. Stores the scale factor in scale 
// if this is the case.
{
	Int		i, j;
	Real	sf = 0.0;
	
	clamped(m1, 1e-8);
	clamped(m2, 1e-8);
	 	
	for (i = 0; i < m1.Rows(); i++)
		for (j = 0; j < m1.Cols(); j++)
			if (abs(m1[i][j]) < 1e-8)
				if (abs(m2[i][j]) > 1e-8)
					return(false);
				else
					;	// 0/0: the scale factor could be anything, so ignore it.
			else if (sf == 0)
				sf = m2[i][j] / m1[i][j];
			else if (fabs(m2[i][j] / m1[i][j] - sf) >= 1e-5)
				return(false);
	
	scale = sf;
	
	return(true);
}

Bool IsTensor(Matd M, Matd &R, Matd &T, Bool &split)
//	Checks whether a matrix can be split up into before & after matrices of
//	half the size. I.e., whether we can replace Mv with RmT, where m is v packed into
//	a square matrix.
//	This corresponds to forming a 2D rule by running 1D rules across the data and then
//	down the data. Rt = T in the case where the same 1D rule is used both horizontally
//	and vertically.
//	As well as saving storage space, using the tensor version of the rule saves operations.
//	E.g. for a 4 -> 16 element 2D push, Mx requires 64*, 48+, whereas RxT requires 48*, 24+.
{
	Int s = M.Cols(), i, j;
	Real d;
	Matd	F, Z;
	
	if ((M.Rows() != M.Cols()) || sqrt(s) != (Int) sqrt(s))
		return(false);
	
	s = sqrt(s);
	
	R.SetSize(s, s);
	T.SetSize(s, s);
	Z.SetSize(s, s);
	Z = vl_zero;
	
	for (i = 0; i < s; i++)
		for (j = 0; j < s; j++)
		{
			F = sub(M, i * s, j * s, s, s);
			clamped(F, 1e-8);
			if (F != Z)
				break;
		}
//	cout << "F = " << F << endl;
	for (i = 0; i < s; i++)
		for (j = 0; j < s; j++)
			if (!MatchMatrix(sub(F, s, s), sub(M, i * s, j * s, s, s), R[i][j]))
				return(false);
				
	T = trans(F);
 
//	cout << "T[0][0] = " << T[0][0] << endl;
	if (abs(T[0][0]) > 1e-8 && abs(R[0][0]) > 1e-8)
	{
		d = sqrt(abs(T[0][0]) / abs(R[0][0]));
		T /= d;
		R *= d;		
	}
				
	return(true);
}

